#include "def_vst.h"
#include "ufv_statistik.h"

/*----------------------------------------------------------------------*/
int tokens_of(char *in, char *out[])
{   NAME("tokens_of()");
    int tc=0,i=0;
    char *cp;
    bool within_token = FALSE;

    cp = in;
    do
    {    if(within_token && (*cp==' ' || *cp=='\t'))
         {   within_token = FALSE;
             *cp=0;
         }
         else if(!within_token && *cp!=' ' && *cp!='\t')
         {    within_token = TRUE;
              out[tc++]=cp;
         }
    }    while(i++, *++cp);
    return tc;
}


int main(int argc, char *argv[])
{
    int      i,n,
             nf;                     /* # floats                    */
    char     *line_str,
             *cp;
    float    ufv[MAX_N],             /* Eingabevektor               */
             min[MAX_N],             /* Minimum                     */
             max[MAX_N],             /* Maximum                     */
             mean[MAX_N],            /* Mittelwert                  */
             vari[MAX_N],            /* Varianz                     */
             negdev[MAX_N],          /* negativ deviation           */
             posdev[MAX_N],          /* positiv deviation           */
             extmean[MAX_N];         /* previously calculated means */

    static   double   mom1[MAX_N],   /* Summe der 1. Momente        */
                      mom2[MAX_N];   /* Summe der 2. Momente        */
    long     vec_cnt=0;              /* Anzahl der ufv's            */
    long     nneg=0, npos=0;
    bool     headline;               /* Kopfzeile ausgeben          */

    
    #define  MAXLL 2048
    char     *meanfn,
             *xfn,
             line[MAXLL];
    char     *token[MAXLL/2];

    FILE     *meanfd,
             *xfd;

    struct index_range
    {   int first;
        int last;
        struct index_range *next;
    };


    struct index_range  *first_range=0, *range, *prev_range=0;
    
    static char *synopsis =
"~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n"
"                                       V1.4, 14/12/2006,  V.Strom  \n"
"                                                                   \n"
" usage: ufv_statistik  n  line  <ufv  >stat                        \n"
"                                                                   \n"
" <ufv  unlabled feature vectors              (float[])             \n"
" n     vector length (nuber of floats)       (int)                 \n"
" line  format of output line    (e.g. 'ims') (string)              \n"
"       i  feature index                                            \n"
"       m  mean                                                     \n"
"       v  variance                                                 \n"
"       s  standard deviation                                       \n"
"       x  minimum                                                  \n"
"       X  maximum                                                  \n"
"       n  mean of negative deviation of mean                       \n"
"       p  mean of positive deviation of mean                       \n"
"       l  mean - mean of negative deviation of mean                \n"
"       h  mean of positive deviation of mean -mean                 \n"
" >stat one ASCII line per feature according to line string.        \n"
"                                                                   \n"
" OPTIONS                                                           \n"
"                                                                   \n"
" -L       output headline first                                    \n"
" -m mean  file containing n previously determined mean values as   \n"
"          ASCII floats, one per line (run ufv_statistik n m)       \n"
"          to be able to calculate statistics n,p,l and h.          \n"
" -x xfn   file containing a list of vector indices which shall     \n"
"          be excluded from the statistic.  Indices should be       \n"
"          separated by white space, ranges are allowed, too,       \n"
"          e.g. '0-13 15 91-99'  Note that counting starts with 0,  \n"
"          ranges are inclusive and no white space within a range.  \n"
"                                                                   \n"
" NOTE  any of [nplh] require to run 'ufv_statistik n m', store     \n"
"       the means in a file, and give it to -m.                     \n"
"                                                                   \n"
"~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n";
    ARGS(2,"L","mx");

    IARG(1,nf);
    if(nf<1 || nf >MAX_N)
        ERROR1("n is out of range [1:%d]",MAX_N);

    SARG(2,line_str);

    BOPT('L',headline);
    SOPT('m',meanfn,NULL);
    SOPT('x',xfn,NULL);

    for(i=0; i<nf; i++)
    {   min[i] =  FLT_MAX;
        max[i] = -FLT_MAX;
    }
    /*======================================================*/
    if(meanfn)
    {   FOPEN(meanfd, meanfn, "r");
        i = 0;
        while(fgets(line, MAXLL, meanfd) && i<MAX_N)
        {   if(sscanf(line, "%f", &extmean[i])!=1)
                ERROR2("cannot scan a float in file %s, line %d",
                       meanfn, i+1);
            DMESSAGE2("extmean[%d] = %f", i, extmean[i]);
            i++;
        }
        if(i!=nf)
            ERROR3("number of lines in file %s (%d) does not match "
                   "number of features (%d)", meanfn, i, nf);
    }
    /*======================================================*/
    if(xfn)
    {   FOPEN(xfd, xfn, "r");
        while(fgets(line, MAXLL, xfd))
        {   n = tokens_of(line, token);
            for(i=0; i<n; i++)
            {   ALLOCATE(range, struct index_range, 1);
                if(sscanf(token[i], "%d-%d", &range->first, &range->last)==2)
                {   if(range->first<0 || range->first>range->last)
                        ERROR3("illegal range %d-%d in %s", range->first,
                                                            range->last, xfn);
                }
                else if(sscanf(token[i], "%d", &range->first)==1)
                {   range->last = range->first;
                    }
                else
                   ERROR2("format error in file %s: %s",xfn,token[i]);
                if(prev_range)
                {   if(prev_range->last >= range->first)
                        ERROR3("indices in %s must be strictly monotonic "
                               "%d >= %d", xfn, prev_range->last, range->first);
                }
                range->next=0;

                if(!first_range)
                {   first_range = range;
                    prev_range = range;
                }
                else
                {   prev_range->next = range;
                    prev_range = range;
                }
            }
        }
    }
    IF_DEBUG
    {   range=first_range;
        while(range)
        {   fprintf(stderr,"RANGE %d - %d\n", range->first, range->last);
            range=range->next;
        }
    }
    range=first_range;
    /*============ Hauptschleife: ==========================*/
    while(FREAD(ufv, float, EXACTLY, nf, stdin))
    {   if(range)
        {   if(vec_cnt>=range->first && vec_cnt<=range->last)
            {    vec_cnt++;
                 continue;
            }
            if(vec_cnt>range->last)
                range=range->next;
        }
        vec_cnt++;
        for(i=0; i<nf; i++)
        {   mom1[i] += ufv[i];
            mom2[i] += SQR(ufv[i]);
            
            if(ufv[i] > max[i])   max[i] = ufv[i];
            if(ufv[i] < min[i])   min[i] = ufv[i];

            if(meanfn)
            {   if(ufv[i] < extmean[i])
                {   negdev[i] += extmean[i]-ufv[i];
                    nneg++;
                }
                else
                {   posdev[i] += ufv[i]-extmean[i];
                    npos++;
                }
            }
        }
        if(vec_cnt%1000 == 0)
	    DMESSAGE1("%ld vectors read in",vec_cnt);
    }
    /*======================================================*/
    if(vec_cnt == 0)
        ERROR0("input file empty");


    /*======== Mittelwert und Varianz: =====================*/
    for(i=0; i<nf; i++)
    {   mean[i] = mom1[i]/vec_cnt;
        vari[i] = mom2[i]/vec_cnt - SQR(mean[i]);
        if(meanfn)
        {   negdev[i] /= nneg;
            posdev[i] /= npos;
        }
    }

    /*======== Ausgabe =====================================*/
    if(headline)
    {   for(cp=line_str; *cp!='\0'; cp++)
        {    switch(*cp)
             {
                 case 'i': printf("index\t");
                           break;
                 case 'm': printf("mean\t");
                           break;
                 case 'v': printf("vari\t");
                           break;
                 case 's': printf("sigma\t");
                           break;
                 case 'x': printf("min\t");
                           break;
                 case 'X': printf("max\t");
                           break;
                 case 'n': printf("neg\t");
                           break;
                 case 'p': printf("pos\t");
                           break;
                 case 'l': printf("low\t");
                           break;
                 case 'h': printf("high\t");
                           break;
                 default:  ERROR1("illegal character \"%c\" in line string",
                                  *cp);
             }
        }
        putchar('\n');
    }
    for(i=0; i<nf; i++)
    {   for(cp=line_str; *cp!='\0'; cp++)
        {    switch(*cp)
             {
                 case 'i': printf("%2d\t", i);
                           break;
                 case 'm': printf("%g\t", mean[i]);
                           break;
                 case 'v': printf("%g\t", vari[i]);
                           break;
                 case 's': printf("%g\t", (float)sqrt(vari[i]));
                           break;
                 case 'x': printf("%g\t", min[i]);
                           break;
                 case 'X': printf("%g\t", max[i]);
                           break;
                 case 'n': if(!meanfn)
                               ERROR0("for calculating l I need a mean file");
                           printf("%g\t", negdev[i]);
                           break;
                 case 'p': if(!meanfn)
                               ERROR0("for calculating h I need a mean file");
                           printf("%g\t", posdev[i]);
                           break;
                 case 'l': if(!meanfn)
                               ERROR0("for calculating l I need a mean file");
                           printf("%g\t", extmean[i]-negdev[i]);
                           break;
                 case 'h': if(!meanfn)
                               ERROR0("for calculating h I need a mean file");
                           printf("%g\t", extmean[i]+posdev[i]);
                           break;
                 default:  ERROR1("illegal character \"%c\" in line string", 
                                  *cp);
             }
        }
        putchar('\n');
    }

    return 0;
}

